/**
 * @file
 * @brief CAN driver thread.
 */

#include <sta/config.hpp>
#ifdef STA_CAN_BUS_ENABLE

#include <sta/debug/assert.hpp>
#include <sta/debug/debug.hpp>
#include <sta/bus/can/subscribable.hpp>
#include <sta/lang.hpp>
#include <sta/proto/isotp/transmitter.hpp>
#include <sta/proto/isotp/receiver.hpp>
#include <sta/rtos/defs.hpp>
#include <sta/rtos/system/events.hpp>
#include <sta/devices/stm32/hal.hpp>
#include <sta/rtos/system/events.hpp>
#include <sta/rtos/system/can_bus.hpp>

#include <cmsis_os2.h>
#include <FreeRTOS.h>

#include <cstring>

namespace debug
{
	/**
	 * @brief Output CAN frame ID to UART.
	 *
	 * @param id    Frame ID
	 */
	void printFrameID(const sta::CanFrameId & id)
	{
		STA_DEBUG_PRINT("SID: ");
		STA_DEBUG_PRINTLN(id.sid, sta::IntegerBase::HEX);
		if (id.format == sta::CanIdFormat::EXT)
		{
			STA_DEBUG_PRINT("EID: ");
			STA_DEBUG_PRINTLN(id.eid, sta::IntegerBase::HEX);
		}
	}

	/**
	 * @brief Output CAN frame payload to UART.
	 *
	 * @param payload   Payload buffer
	 * @param size      Payload size
	 */
	void printPayloadHex(const uint8_t * payload, uint8_t size)
	{
		// Write frame payload to DebugSerial
		STA_DEBUG_PRINT("payload: ");
		for (uint8_t i = 0; i < size; ++i)
		{
			STA_DEBUG_PRINT(payload[i], sta::IntegerBase::HEX);
			STA_DEBUG_PRINT(' ');
		}
		STA_DEBUG_PRINTLN();
	}
} // namespace debug

namespace sta
{
	AlpakaCanBus::AlpakaCanBus(CanController * controller, TimeMsFn timeMs, SysMsgHandler sysMsgHandler, DataMsgHandler dataMsgHandler)
		: controller_{controller}, tx_{controller, timeMs}, rx_{controller, timeMs}, handleSysMsg_{sysMsgHandler}, handleDataMsg_{dataMsgHandler}
	{
		STA_ASSERT(handleSysMsg_ != nullptr);
		STA_ASSERT(handleDataMsg_ != nullptr);

		setupSubscriptions();
	}


	void AlpakaCanBus::send(const CanSysMsg & msg)
	{
		CanTxHeader header;
		header.id.format     = static_cast<CanIdFormat>(msg.header.format);
		header.id.sid        = msg.header.sid & STA_CAN_SID_SYS_BITS;
		header.id.eid        = msg.header.eid;
		header.payloadLength = msg.header.payloadLength;

		controller_->sendFrame(header, msg.payload);
	}

	void AlpakaCanBus::send(const CanDataMsg & msg)
	{
		CanFrameId frameID;
		frameID.format = static_cast<CanIdFormat>(msg.header.format);
		frameID.sid    = msg.header.sid & ~STA_CAN_SID_SYS_BITS;
		frameID.eid    = msg.header.eid;

		// Start transmission via ISO-TP
		tx_.send(frameID, msg.payload, msg.header.payloadLength);
	}


	void AlpakaCanBus::processTx()
	{
		tx_.process();
	}


	void AlpakaCanBus::processRx()
	{
		for (auto fifo : controller_->getPendingRxFifos())
		{
			CanRxHeader header;
			uint8_t payload[STA_RTOS_CAN_BUS_MAX_PAYLOAD_SIZE];

			if (controller_->receiveFrame(fifo, &header, payload))
			{
				//debug::displayFrameUART(frame);

				// Forward frame to callback
				switch (fifo)
				{
				case FIFO_SYS:
					handleSysMsg_(header, payload);
					break;

				case FIFO_DATA:
					receiveDataFrame(header, payload);
					break;

				default:
					STA_ASSERT(false);
				}
			}
		}
	}

	void AlpakaCanBus::showStatistics()
	{
		STA_DEBUG_PRINTLN();
		STA_DEBUG_PRINTLN("# ######################");
		STA_DEBUG_PRINTLN("# ## ISOTP statistics ##");
		STA_DEBUG_PRINTLN("# ######################");
		STA_DEBUG_PRINTLN("#");

		const auto & txStats = tx_.stats();
		STA_DEBUG_PRINTLN("# Transmitter");
		STA_DEBUG_PRINT("#     messages: ");
		STA_DEBUG_PRINTLN(txStats.messages);
		STA_DEBUG_PRINT("#     blocks: ");
		STA_DEBUG_PRINTLN(txStats.blocks);
		STA_DEBUG_PRINT("#     frames: ");
		STA_DEBUG_PRINTLN(txStats.frames);
		STA_DEBUG_PRINT("#     timeouts: ");
		STA_DEBUG_PRINTLN(txStats.timeouts);
		STA_DEBUG_PRINTLN("#");

		const auto & rxStats = rx_.stats();
		STA_DEBUG_PRINTLN("# Receiver");
		STA_DEBUG_PRINT("#     messages: ");
		STA_DEBUG_PRINTLN(rxStats.messages);
		STA_DEBUG_PRINT("#     blocks: ");
		STA_DEBUG_PRINTLN(rxStats.blocks);
		STA_DEBUG_PRINT("#     frames: ");
		STA_DEBUG_PRINTLN(rxStats.frames);
		STA_DEBUG_PRINT("#     timeouts: ");
		STA_DEBUG_PRINTLN(rxStats.timeouts);
		STA_DEBUG_PRINT("#     flow control errors: ");
		STA_DEBUG_PRINTLN(rxStats.flowErrors);
		STA_DEBUG_PRINT("#     overflows: ");
		STA_DEBUG_PRINTLN(rxStats.overflows);
		STA_DEBUG_PRINTLN();
	}


	void AlpakaCanBus::setupSubscriptions()
	{
		// Make sure to receive all messages
		CanFilter filter;

		// Clear previous subscriptions
		controller_->clearFilters();


		// All bits except [0:1] of the SID must be zero for system messages
		filter.obj  = {0, 0};
		filter.mask = {~STA_CAN_SID_SYS_BITS, 0};
		filter.type = sta::CanFilterIdFormat::ANY;
		filter.fifo = FIFO_SYS;
		controller_->configureFilter(FIFO_SYS, filter, true);

		// TODO Limit which data messages are received
		// Bits [0:1] of the SID must be zero for data messages
		filter.obj  = {0, 0};
		filter.mask = {STA_CAN_SID_SYS_BITS, 0};
		filter.type = sta::CanFilterIdFormat::ANY;
		filter.fifo = FIFO_DATA;
		controller_->configureFilter(FIFO_DATA, filter, true);
	}

	void AlpakaCanBus::receiveDataFrame(const CanRxHeader & header, const uint8_t * payload)
	{
		// Write frame payload to DebugSerial
		STA_DEBUG_PRINTLN("[event] RX data frame");
		debug::printPayloadHex(payload, header.payloadLength);

		// Process RX frame
		auto handle = rx_.processFrame(header, payload);

		if (handle != IsotpReceiver::INVALID_HANDLE)
		{
			// Get message if completed
			IsotpMessage msg;
			if (rx_.getMessage(handle, &msg))
			{
				handleDataMsg_(msg);
			}

			// Handle FC responses
			rx_.processFC(handle);
		}

		// Process TX frame
		tx_.processFrame(header, payload);
	}
} // namespace sta

#endif // STA_CAN_BUS_ENABLE
